"""
Contains the tuning parameter selection algorithm
"""

import numpy as np
import pandas as pd
from scipy.stats import skew, skewnorm
from sklearn.utils.parallel import Parallel, delayed
import matplotlib.pyplot as plt

from ._utils import stat_ksample, stat_two_sample, stat_normality_test
from ._cv_functions import cv_ksample, cv_twosample, cv_normality


def _objective_one_sample(
    alternative,
    delta,
    delta_dim,
    h,
    mean_dat,
    n,
    num_iter,
    quantile,
    rep_values,
    s_dat,
    skew_data,
    random_state,
    n_jobs=1,
):
    """
    Objective function using using the best
    h is chosen for one sample test

    Parameters
    ----------
        alternative : str
            family of alternative chosen for selecting h,
            must be one of "mean", "variance" and "skewness".

        delta : numpy.ndarray
            Array of parameter values indicating chosen alternatives.

        delta_dim : int, numpy.ndarray
            Array of coefficient of alternative with respect
            to each dimension.

        h : float
            Bandwidth for the kernel function.

        mean_dat : numpy.ndarray
            Means of the multivariate distribution to be used
            for determining the best h.

        n : int
            Number of observations in set of samples.

        num_iter : int
            The number of iterations to use for critical value estimation.

        quantile : float
            Quantile to use for critical value estimation.

        rep_values : int
            Number of the bootstrap replication.

        s_dat : numpy.ndarray
            Variances of the multivariate distribution to be used
            for determining the best h.

        skew_data : numpy.ndarray
            Skewness of the multivariate distribution to be used
            for determining the best h.

        random_state : int, None, optional.
            Seed for random number generation. Defaults to None.

        n_jobs : int, optional
            n_jobs specifies the maximum number of concurrently
            running workers. If 1 is given, no joblib parallelism
            is used at all, which is useful for debugging. For more
            information on joblib n_jobs refer to -
            https://joblib.readthedocs.io/en/latest/generated/joblib.Parallel.html.

    Returns
    -------
        List containing rep_values, delta, h and boolean
        representing the rejection/acceptance of null hypothesis.
    """
    dk = delta_dim * delta
    if alternative == "location":
        mean_tilde = mean_dat + dk
        s_tilde = s_dat
        skew_tilde = skew_data
    elif alternative == "scale":
        mean_tilde = mean_dat
        s_tilde = s_dat * dk
        skew_tilde = skew_data
    elif alternative == "skewness":
        mean_tilde = mean_dat
        skew_tilde = skew_data + dk
        s_tilde = s_dat

    if isinstance(random_state, int):
        random_state = random_state + int(rep_values)
    xnew = skewnorm.rvs(
        size=(n, len(mean_dat)),
        loc=mean_tilde,
        scale=s_tilde,
        a=skew_tilde,
        random_state=random_state,
    )

    statistic = stat_normality_test(xnew, h, np.array([mean_dat]), np.diag(s_dat))
    cv = cv_normality(
        n,
        h,
        np.array([mean_dat]),
        np.diag(s_dat),
        num_iter,
        quantile,
        random_state=random_state,
        n_jobs=n_jobs,
    )
    h0 = statistic < cv
    return [rep_values, delta, h, h0[0]]


def _objective_two_sample(
    alternative,
    b,
    delta,
    delta_dim,
    h,
    m,
    mean_dat,
    method,
    n,
    num_iter,
    pooled,
    quantile,
    rep_values,
    s_dat,
    skew_data,
    d,
    random_state,
    n_jobs=1,
):
    """
    Objective function using using the best
    h is chosen for two sample test.

    Parameters
    ----------
        alternative : str
            family of alternative chosen for selecting h,
            must be one of "mean", "variance" and "skewness".

        b : float
            The size of the subsamples used in the subsampling algorithm.

        delta : numpy.ndarray
            Array of parameter values indicating chosen alternatives.

        delta_dim : int, numpy.ndarray
            Array of coefficient of alternative with respect
            to each dimension.

        h : float
            Bandwidth for the kernel function.

        m : int
            Number of observations in second set of samples.

        mean_dat : numpy.ndarray
            Means of the multivariate distribution to be used
            for determining the best h.

        method : str
            the method used for critical value estimation,
            must be one of "subsampling", "bootstrap", or "permutation".

        n : int
            Number of observations in first set of samples.

        num_iter : int
            The number of iterations to use for critical value estimation.

        pooled : numpy.ndarray
            Observations in first set and second
            set of samples combined together retaining the number of columns.

        quantile : float
            Quantile to use for critical value estimation.

        rep_values : int
            Number of the bootstrap replication.

        s_dat : numpy.ndarray
            Variances of the multivariate distribution to be used
            for determining the best h.

        skew_data : numpy.ndarray
            Skewness of the multivariate distribution to be used
            for determining the best h.

        d : int
            Dimension of the data.

        random_state : int, None, optional.
            Seed for random number generation. Defaults to None.

        n_jobs : int, optional
            n_jobs specifies the maximum number of concurrently
            running workers. If 1 is given, no joblib parallelism
            is used at all, which is useful for debugging. For more
            information on joblib n_jobs refer to -
            https://joblib.readthedocs.io/en/latest/generated/joblib.Parallel.html.

    Returns
    -------
        List containing rep_values, delta, h and boolean
        representing the rejection/acceptance of null hypothesis.
    """

    dk = delta_dim * delta
    if alternative == "location":
        mean_tilde = mean_dat + dk
        s_tilde = s_dat
        skew_tilde = skew_data
    elif alternative == "scale":
        mean_tilde = mean_dat
        s_tilde = s_dat * dk
        skew_tilde = skew_data
    elif alternative == "skewness":
        mean_tilde = mean_dat
        skew_tilde = skew_data + dk
        s_tilde = s_dat

    if isinstance(random_state, (int, np.int_)):
        random_state = random_state + int(rep_values)

    xnew = skewnorm.rvs(
        size=(n, len(mean_dat)),
        loc=mean_dat,
        scale=s_dat,
        a=skew_data,
        random_state=np.random.default_rng(random_state),
    )
    ynew = skewnorm.rvs(
        size=(m, len(mean_dat)),
        loc=mean_tilde,
        scale=s_tilde,
        a=skew_tilde,
        random_state=np.random.default_rng(random_state),
    )

    statistic = stat_two_sample(xnew, ynew, h, np.repeat(0, d), np.eye(d))
    cv = cv_twosample(
        num_iter,
        quantile,
        pooled,
        n,
        m,
        h,
        method,
        b,
        random_state=random_state,
        n_jobs=n_jobs,
    )
    h0 = statistic[:2] < cv
    return [rep_values, delta, h, h0[0]]


def _objective_k_sample(
    alternative,
    num_iter,
    b,
    delta,
    delta_dim,
    h,
    k,
    mean_dat,
    method,
    n,
    quantile,
    rep_values,
    s_dat,
    skew_data,
    random_state,
    n_jobs=1,
):
    """
    Objective function using using the best
    h is chosen for k-sample test.

    Parameters
    ----------
        alternative : str
            family of alternative chosen for selecting h,
            must be one of "mean", "variance" and "skewness".

        num_iter : int
            The number of iterations to use for critical value estimation.

        b : float
            The size of the subsamples used in the subsampling algorithm.

        delta : numpy.ndarray
            Array of parameter values indicating chosen alternatives.

        delta_dim : int, numpy.ndarray
            Array of coefficient of alternative with respect
            to each dimension.

        h : float
            Bandwidth for the kernel function.

        k : int
            Number of classes (or groups) in the data.

        mean_dat : numpy.ndarray
            Means of the multivariate distribution to be used
            for determining the best h.

        method : str
            the method used for critical value estimation,
            must be one of "subsampling", "bootstrap", or "permutation".

        n : int
            Number of total observations in the provided data.

        quantile : float
            Quantile to use for critical value estimation.

        rep_values : int
            Number of the bootstrap replication.

        s_dat : numpy.ndarray
            Variances of the multivariate distribution to be used
            for determining the best h.

        skew_data : numpy.ndarray
            Skewness of the multivariate distribution to be used
            for determining the best h.

        random_state : int, None, optional.
            Seed for random number generation. Defaults to None.

        n_jobs : int, optional
            n_jobs specifies the maximum number of concurrently
            running workers. If 1 is given, no joblib parallelism
            is used at all, which is useful for debugging. For more
            information on joblib n_jobs refer to -
            https://joblib.readthedocs.io/en/latest/generated/joblib.Parallel.html.

    Returns
    -------
        List containing rep_values, delta, h and boolean
        representing the rejection/acceptance of null hypothesis.
    """
    dk = delta_dim * delta
    if alternative == "location":
        mean_tilde = mean_dat + dk
        s_tilde = s_dat
        skew_tilde = skew_data
    elif alternative == "scale":
        mean_tilde = mean_dat
        s_tilde = s_dat * dk
        skew_tilde = skew_data
    elif alternative == "skewness":
        mean_tilde = mean_dat
        skew_tilde = skew_data + dk
        s_tilde = s_dat

    if isinstance(random_state, (int, np.int_)):
        random_state = random_state + int(rep_values)

    nk = round(n / k)
    xnew = skewnorm.rvs(
        size=(nk * (k - 1), len(mean_dat)),
        loc=mean_dat,
        scale=s_dat,
        a=skew_data,
        random_state=random_state,
    )
    xk = skewnorm.rvs(
        size=(nk, len(mean_dat)),
        loc=mean_tilde,
        scale=s_tilde,
        a=skew_tilde,
        random_state=random_state,
    )

    xnew = np.concatenate((xnew, xk), axis=0)
    ynew = np.repeat(np.arange(1, k + 1), repeats=nk)
    statistic = stat_ksample(xnew, ynew, h)
    cv = cv_ksample(
        xnew,
        ynew,
        h,
        num_iter,
        b,
        quantile,
        method,
        random_state=random_state,
        n_jobs=n_jobs,
    )
    h0 = statistic[:2] < cv
    return [rep_values, delta, h, h0[0]]


def select_h(
    x,
    y=None,
    alternative="location",
    method="subsampling",
    b=0.8,
    num_iter=150,
    delta_dim=1,
    delta=None,
    h_values=None,
    n_rep=50,
    n_jobs=8,
    quantile=0.95,
    k_threshold=10,
    power_plot=False,
    random_state=None,
):
    """
    This function computes the kernel bandwidth of the Gaussian kernel
    for the one sample, two-sample and k-sample kernel-based quadratic
    distance (KBQD) tests.

    The function performs the selection of the optimal value for the tuning
    parameter h of the normal kernel function, for the two-sample and k-sample
    KBQD tests. It performs a small simulation study, generating samples according
    to the family of a specified alternative, for the chosen values
    of h_values and delta.

    Parameters
    ----------
        x : numpy.ndarray or pandas.DataFrame
            Data set of observations from X.

        y : numpy.ndarray or pandas.DataFrame, optional
            Data set of observations from Y for two sample test
            or set of labels in case of k-sample test.

        alternative : str, optional
            Family of alternative chosen for selecting h,
            must be one of "location", "scale" and "skewness".
            Defaults to "location".

        method : str, optional.
            The method used for critical value estimation,
            must be one of "subsampling", "bootstrap", or "permutation".
            Defaults to "subsampling".

        b : float, optional.
            The size of the subsamples used in the subsampling algorithm.
            Defaults to 0.8 i.e. `0.8N` samples are used, where `N`
            represents the total sample size.

        num_iter : int, optional.
            The number of iterations to use for critical value estimation.
            Defaults to 150.

        delta_dim : int, numpy.ndarray, optional.
            Array of coefficient of alternative with respect to each dimension.
            Defaults to 1.

        delta : numpy.ndarray, optional.
            Array of parameter values indicating chosen alternatives.
            Defaults to None.

        h_values : numpy.ndarray, optional.
            Values of the tuning parameter used for the selection.
            Defaults to None.

        n_rep : int, optional. Defaults to 50.
            Number of bootstrap replications.

        n_jobs : int, optional.
            n_jobs specifies the maximum number of concurrently running workers.
            If 1 is given, no joblib parallelism is used at all,
            which is useful for debugging. For more information on joblib n_jobs
            refer to - https://joblib.readthedocs.io/en/latest/generated/joblib.Parallel.html.
            Defaults to 8.

        quantile : float, optional.
            Quantile to use for critical value estimation. Defaults to 0.95.

        k_threshold : int.
            Maximum number of groups allowed. Defaults to 10.

        power_plot : boolean, optional.
            If True, plot is displayed the plot of power for
            values in h\\_values and delta. Defaults to False.

        random_state : int, None, optional.
            Seed for random number generation. Defaults to None.

    Returns
    -------
        h : float
            The selected value of tuning parameter h.

        h vs Power table : pandas.DataFrame
            A table containing the h, delta and corresponding powers.

    References
    -----------
        Markatou M., Saraceno G., Chen Y. (2023). “Two- and k-Sample Tests
        Based on Quadratic Distances. ”Manuscript, (Department of Biostatistics,
        University at Buffalo).

    Examples
    --------
    >>> import numpy as np
    >>> from QuadratiK.kernel_test import select_h
    >>> np.random.seed(42)
    >>> X = np.random.randn(200, 2)
    >>> np.random.seed(42)
    >>> y = np.random.randint(0, 2, 200)
    >>> h_selected, all_values, power_plot = select_h(
    ...    X, y, alternative='location', power_plot=True, random_state=42)
    >>> print("Selected h is: ", h_selected)
    ... Selected h is:  1.2
    """
    if not isinstance(random_state, (int, np.int_, type(None))):
        raise ValueError("Please specify a integer or None random_state")

    if isinstance(x, np.ndarray):
        if x.ndim == 1:
            x = np.array(x).reshape(-1, 1)
    elif isinstance(x, pd.DataFrame):
        x = x.to_numpy()
    else:
        raise TypeError("x must be a numpy array or a pandas DataFrame")

    if y is not None:
        if isinstance(y, (list, np.ndarray, pd.Series)):
            if y.ndim > 1:
                pass
            else:
                y = np.array(y).reshape(-1, 1)
        elif isinstance(y, pd.DataFrame):
            y = y.values
        else:
            raise TypeError(
                "y must be a list, numpy.ndarray, \
                    or a pandas DataFrame with one column (for k-sample case) or more"
            )
    else:
        pass

    if h_values is None:
        h_values = np.round(np.arange(0.4, 3.4, 0.4), 2)

    if delta is None:
        if alternative == "location":
            delta = np.array([0.2, 0.3, 0.4])
        elif alternative == "scale":
            delta = np.array([1.1, 1.3, 1.5])
        elif alternative == "skewness":
            delta = np.array([0.2, 0.3, 0.6])

    if (alternative is None) or (alternative not in ["location", "scale", "skewness"]):
        raise ValueError(
            "Please specify alternative from 'location', 'scale' or 'skewness'"
        )

    if y is not None:
        sizes = np.unique(y, return_counts=True)[1]
    else:
        sizes = []

    rep_values = np.arange(0, n_rep, 1)
    k = len(sizes)

    if k <= 1:
        n, d = x.shape
        if isinstance(delta_dim, (int, float)):
            delta_dim = np.ones(d)
        else:
            if not isinstance(delta_dim, (int, float)) or len(delta_dim) != d:
                raise ValueError(
                    "delta_dim must be 1 or a numeric Array of \
                        length equal to the number of columns of pooled."
                )

        mean_dat = np.mean(x, axis=0)
        s_dat = np.diag(np.cov(x, rowvar=False).reshape(x.shape[1], x.shape[1]))
        skew_data = skew(x)
        all_parameters = np.array(np.meshgrid(h_values, delta, rep_values)).T.reshape(
            -1, 3
        )

        all_results = {}
        for delta_val in delta:
            parameters = all_parameters[all_parameters[:, 1] == delta_val]
            results = Parallel(n_jobs=n_jobs)(
                delayed(_objective_one_sample)(
                    alternative,
                    param[1],
                    delta_dim,
                    param[0],
                    mean_dat,
                    n,
                    num_iter,
                    quantile,
                    param[2],
                    s_dat,
                    skew_data,
                    random_state,
                    1,
                )
                for param in parameters
            )
            results = pd.DataFrame(results, columns=["rep", "delta", "h", "score"])
            results["score"] = 1 - results["score"]
            results_mean = (
                results.groupby(["h", "delta"]).agg({"score": "mean"}).reset_index()
            )
            results_mean.columns = ["h", "delta", "power"]
            results_mean = results_mean.sort_values(by=["delta", "h"])
            all_results[delta_val] = results_mean
            min_h_power_gt_05 = results_mean[results_mean["power"] >= 0.5]
            if not min_h_power_gt_05.empty:
                min_h = results_mean[results_mean["power"] >= 0.50].iloc[0]["h"]
                break

    elif k > k_threshold:
        if x.shape[1] != y.shape[1]:
            raise ValueError("'x' and 'y' must have the same number of columns")

        n = x.shape[0]
        m = y.shape[0]
        pooled = np.concatenate((x, y), axis=0)
        d = pooled.shape[1]

        if isinstance(delta_dim, (int, float)):
            delta_dim = np.ones(d)
        else:
            if not isinstance(delta_dim, (int, float)) or len(delta_dim) != d:
                raise ValueError(
                    "delta_dim must be 1 or a numeric Array of \
                        length equal to the number of columns of pooled."
                )

        mean_dat = np.mean(pooled, axis=0)
        s_dat = np.diag(
            np.cov(pooled, rowvar=False).reshape(pooled.shape[1], pooled.shape[1])
        )
        skew_data = skew(pooled)
        all_parameters = np.array(np.meshgrid(h_values, delta, rep_values)).T.reshape(
            -1, 3
        )

        all_results = {}
        for delta_val in delta:
            parameters = all_parameters[all_parameters[:, 1] == delta_val]
            results = Parallel(n_jobs=n_jobs)(
                delayed(_objective_two_sample)(
                    alternative,
                    b,
                    param[1],
                    delta_dim,
                    param[0],
                    m,
                    mean_dat,
                    method,
                    n,
                    num_iter,
                    pooled,
                    quantile,
                    param[2],
                    s_dat,
                    skew_data,
                    d,
                    random_state,
                    1,
                )
                for param in parameters
            )
            results = pd.DataFrame(results, columns=["rep", "delta", "h", "score"])
            results["score"] = 1 - results["score"]
            results_mean = (
                results.groupby(["h", "delta"]).agg({"score": "mean"}).reset_index()
            )
            results_mean.columns = ["h", "delta", "power"]
            results_mean = results_mean.sort_values(by=["delta", "h"])
            all_results[delta_val] = results_mean
            min_h_power_gt_05 = results_mean[results_mean["power"] >= 0.5]
            if not min_h_power_gt_05.empty:
                min_h = results_mean[results_mean["power"] >= 0.50].iloc[0]["h"]
                break
    else:
        n, d = x.shape
        if isinstance(delta_dim, (int, float)):
            delta_dim = np.ones(d)
        else:
            if not isinstance(delta_dim, (int, float)) or len(delta_dim) != d:
                raise ValueError(
                    "delta_dim must be 1 or a numeric Array of \
                        length equal to the number of columns of pooled."
                )

        mean_dat = np.mean(x, axis=0)
        s_dat = np.diag(np.cov(x, rowvar=False).reshape(x.shape[1], x.shape[1]))
        skew_data = skew(x)

        all_parameters = np.array(np.meshgrid(h_values, delta, rep_values)).T.reshape(
            -1, 3
        )

        all_results = {}
        for delta_val in delta:
            parameters = all_parameters[all_parameters[:, 1] == delta_val]
            results = Parallel(n_jobs=n_jobs)(
                delayed(_objective_k_sample)(
                    alternative,
                    num_iter,
                    b,
                    param[1],
                    delta_dim,
                    param[0],
                    k,
                    mean_dat,
                    method,
                    n,
                    quantile,
                    param[2],
                    s_dat,
                    skew_data,
                    random_state,
                    1,
                )
                for param in parameters
            )

            results_df = pd.DataFrame(results, columns=["rep", "delta", "h", "score"])
            results_df["score"] = 1 - results_df["score"]
            results_mean = (
                results_df.groupby(["h", "delta"]).agg({"score": "mean"}).reset_index()
            )
            results_mean.columns = ["h", "delta", "power"]
            results_mean = results_mean.sort_values(by=["delta", "h"])
            all_results[delta_val] = results_mean
            min_h_power_gt_05 = results_mean[results_mean["power"] >= 0.5]
            if not min_h_power_gt_05.empty:
                min_h = results_mean[results_mean["power"] >= 0.50].iloc[0]["h"]
                break

    all_results = pd.concat(all_results.values())
    if "min_h" not in locals():
        min_h = results_mean.loc[results_mean["power"].idxmax()]["h"]

    if power_plot:
        groups = all_results.groupby("delta")
        figure = plt.figure(figsize=(8, 4))
        for delta, group in groups:
            plt.plot(
                group["h"],
                group["power"],
                marker="o",
                linestyle="-",
                label=f"delta={round(delta, 3)}",
            )
        plt.xlabel("h")
        plt.ylabel("Power")
        plt.title("h vs Power for different delta")
        plt.legend()
        plt.close()
        return (min_h, all_results, figure)

    return (min_h, all_results)



"""
Critical value functions for Kernel Test
"""

from sklearn.utils.parallel import Parallel, delayed
import numpy as np
from ._utils import (
    normal_cv_helper,
    bootstrap_helper_twosample,
    permutation_helper_twosample,
    subsampling_helper_twosample,
    bootstrap_helper_ksample,
    subsampling_helper_ksample,
    permutation_helper_ksample,
)


def cv_twosample(
    num_iter,
    quantile,
    data_pool,
    size_x,
    size_y,
    h,
    method,
    b=0.9,
    random_state=None,
    n_jobs=8,
):
    """
    This function computes the critical value for two-sample kernel tests with
    centered Gaussian kernel using one of three methods: bootstrap, permutation, or subsampling.

    Parameters
    ----------
        num_iter : int
            The number of iterations to use for critical value estimation.

        quantile : float
            The quantile to use for critical value estimation.

        data_pool: numpy.ndarray
            ndarray containing the data to be used in the test.

        size_x : int
            The number of rows in the data_pool corresponding to group X.

        size_y : int
            The number of rows in the data_pool corresponding to group Y.

        h : float
            The tuning parameter for the kernel test.

        method : str
            Method to use for computing the critical value
            (one of bootstrap, permutation, or subsampling).

        b : float, optional
            Subsampling block size (only used if method is subsampling).

        random_state : int, None, optional.
            Seed for random number generation. Defaults to None.

        n_jobs : int, optional
            n_jobs specifies the maximum number of concurrently
            running workers. If 1 is given, no joblib parallelism
            is used at all, which is useful for debugging. For more
            information on joblib n_jobs refer to -
            https://joblib.readthedocs.io/en/latest/generated/joblib.Parallel.html.

    Returns
    -------
        critical value : float
            Critical value for the specified dimension, size and quantile.

    References
    -----------
        Markatou Marianthi, Saraceno Giovanni, Chen Yang (2023). “Two- and k-Sample Tests Based on
        Quadratic Distances.” Manuscript, (Department of Biostatistics, University at Buffalo).

    """
    if method == "bootstrap":
        results = Parallel(n_jobs=n_jobs)(
            delayed(bootstrap_helper_twosample)(
                size_x, size_y, h, data_pool, i, random_state
            )
            for i in range(num_iter)
        )
    elif method == "permutation":
        results = Parallel(n_jobs=n_jobs)(
            delayed(permutation_helper_twosample)(
                size_x, size_y, h, data_pool, i, random_state
            )
            for i in range(num_iter)
        )
    elif method == "subsampling":
        results = Parallel(n_jobs=n_jobs)(
            delayed(subsampling_helper_twosample)(
                size_x, size_y, b, h, data_pool, i, random_state
            )
            for i in range(num_iter)
        )
    cv = np.quantile(results, quantile, axis=0)
    return cv


def cv_normality(
    size,
    h,
    mu_hat,
    sigma_hat,
    num_iter=500,
    quantile=0.95,
    random_state=None,
    n_jobs=8,
):
    """
    This function computes the empirical critical value for the
    Normality test based on the KBQD tests using the centered Gaussian kernel.

    For each replication, a sample from the d-dimensional Normal distribution with mean
    vector mu_hat and covariance matrix sigma_hat is generated and the KBQD test U-statistic
    for Normality is computed. After num_iter iterations, the critical value is selected as
    the quantile of the empirical distribution of the computed test statistics.

    Parameters
    ----------
        d : int
            The dimension of generated samples.

        size : int
            The number of observations to be generated.

        h : float
            The concentration parameter for the Gaussian kernel.

        mu_hat : numpy.ndarray
            Mean vector for the reference distribution.

        sigma_hat : numpy.array
            Covariance matrix of the reference distribution.

        num_iter : int, optional
            The number of replications.

        quantile : float, optional
            The quantile of the distribution used to select the critical value.

        random_state : int, None, optional.
            Seed for random number generation. Defaults to None.

        n_jobs : int, optional
            n_jobs specifies the maximum number of concurrently
            running workers. If 1 is given, no joblib parallelism
            is used at all, which is useful for debugging. For more
            information on joblib n_jobs refer to -
            https://joblib.readthedocs.io/en/latest/generated/joblib.Parallel.html.

    Returns
    -------
        critical value : float
            Critical value for the specified dimension, size and quantile.
    """
    results = Parallel(n_jobs=n_jobs)(
        delayed(normal_cv_helper)(size, h, mu_hat, sigma_hat, i, random_state)
        for i in range(num_iter)
    )
    return np.quantile(results, quantile, axis=0)


def cv_ksample(
    x,
    y,
    h,
    num_iter=150,
    b=0.9,
    quantile=0.95,
    method="subsampling",
    random_state=None,
    n_jobs=8,
):
    """
    Compute the critical value for k-sample kernel tests.

    Parameters
    --------------
        x : numpy.ndarray
            Matrix containing the observations to be used in the k-sample test.

        y : numpy.ndarray
            Vector indicating the sample for each observation.

        h : float
            The tuning parameter for the test using the Gaussian kernel.

        num_iter : int, optional. Defaults to 150.
            The number of bootstrap/permutation/subsampling samples to generate.

        b : float, optional. Defaults to 0.9
            The subsampling block size (only used if `method` is "subsampling").

        quantile : float, optional. Defaults to 0.95
            The quantile of the bootstrap/permutation/subsampling distribution
            to use as the critical value.

        method : str
            The method to use for computing the critical value
            (one of "bootstrap", "permutation" or "subsampling").

        random_state : int, None, optional.
            Seed for random number generation. Defaults to None.

        n_jobs : int, optional
            n_jobs specifies the maximum number of concurrently
            running workers. If 1 is given, no joblib parallelism
            is used at all, which is useful for debugging. For more
            information on joblib n_jobs refer to -
            https://joblib.readthedocs.io/en/latest/generated/joblib.Parallel.html.

    Returns
    ---------
        critical value : float
            A vector of two critical values corresponding to different formulations
            of the test statistics.

    References
    -----------
        Markatou Marianthi, Saraceno Giovanni, Chen Yang (2023). “Two- and k-Sample Tests Based on
        Quadratic Distances.” Manuscript, (Department of Biostatistics, University at Buffalo).
    """
    sizes = np.unique(y, return_counts=True)[1]
    n = len(y)
    k = len(sizes)
    cum_size = np.insert(np.cumsum(sizes), 0, 0)

    if method == "bootstrap":
        results = Parallel(n_jobs=n_jobs)(
            delayed(bootstrap_helper_ksample)(
                x, y, k, h, sizes, cum_size, i, random_state
            )
            for i in range(num_iter)
        )
    elif method == "subsampling":
        results = Parallel(n_jobs=n_jobs)(
            delayed(subsampling_helper_ksample)(
                x, y, k, h, sizes, b, cum_size, i, random_state
            )
            for i in range(num_iter)
        )
    elif method == "permutation":
        results = Parallel(n_jobs=n_jobs)(
            delayed(permutation_helper_ksample)(x, y, n, h, i, random_state)
            for i in range(num_iter)
        )

    cv = np.quantile(results, quantile, axis=0)
    return cv



"""
Tests the PKBD class
"""

import unittest
import numpy as np
import pandas as pd
from QuadratiK.spherical_clustering import PKBD


class TestPKBD(unittest.TestCase):

    def test_rpkb(self):
        pkbd = PKBD()
        x_rejvmf = pkbd.rpkb(100, [1, 0, 0], 0.8, "rejvmf", random_state=42)
        x_rejacg = pkbd.rpkb(100, [1, 0, 0], 0.8, "rejacg", random_state=42)

        self.assertEqual(x_rejvmf.shape, (100, 3))
        self.assertEqual(x_rejacg.shape, (100, 3))
        self.assertIsInstance(x_rejvmf, np.ndarray)
        self.assertIsInstance(x_rejacg, np.ndarray)
        self.assertTrue(np.allclose(np.sum(x_rejvmf**2, axis=1), np.ones(100)))
        self.assertTrue(np.allclose(np.sum(x_rejacg**2, axis=1), np.ones(100)))

        with self.assertRaises(ValueError):
            pkbd.rpkb(100, np.array([1, 0, 0]), -1, "rejvmf", random_state=42)

        with self.assertRaises(ValueError):
            pkbd.rpkb(100, np.array([0, 0, 0]), 0.1, "rejvmf", random_state=42)

        with self.assertRaises(ValueError):
            pkbd.rpkb(100, np.array([1, 0, 0]), 0.1, "some method", random_state=42)

        with self.assertRaises(ValueError):
            pkbd.rpkb(100, np.array([1, 0, 0]), 0.1, "rejvmf", random_state=[42])

        with self.assertRaises(NotImplementedError):
            pkbd.rpkb(100, np.array([1, 0, 0]), 0.1, "rejpsaw", random_state=42)

        with self.assertRaises(ValueError):
            pkbd.rpkb(-100, np.array([1, 0, 0]), 0.1, "rejvmf", random_state=42)

    def test_dpkb(self):
        pkbd = PKBD()
        x = pkbd.rpkb(n=100, mu=np.array([1, 1, 1]), rho=0.9)
        density_x = pkbd.dpkb(
            x=pd.DataFrame(x), mu=np.array([1, 1, 1]), rho=0.9, logdens=True
        )
        density_y = pkbd.dpkb(x=pd.DataFrame(x), mu=np.array([1, 1, 1]), rho=0.9)

        self.assertEqual(density_x.shape[0], 100, "Dimensions are equal")
        self.assertEqual(density_y.shape[0], 100, "Dimensions are equal")

        with self.assertRaises(ValueError):
            pkbd.dpkb(x=x, mu=np.array([1]), rho=0.9)

        with self.assertRaises(ValueError):
            x_data = np.array([[1, 1, 1, 1, 1]]).reshape(-1, 1)
            pkbd.dpkb(x=x_data, mu=np.array([1, 1]), rho=0.9)

        with self.assertRaises(ValueError):
            pkbd.dpkb(x=x[:, 1], mu=np.array([1, 1]), rho=0.9)

        with self.assertRaises(ValueError):
            pkbd.dpkb(x=x, mu=np.array([1, 1]), rho=0.9)

        with self.assertRaises(ValueError):
            pkbd.dpkb(x=x, mu=np.array([1, 1, 1]), rho=-0.9)

        with self.assertRaises(ValueError):
            pkbd.dpkb(x=x, mu=np.array([0, 0, 0]), rho=0.9)



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import os
import sys

basedir = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", ".."))
sys.path.insert(0, basedir)

project = "QuadratiK"
copyright = (
    "2023, Giovanni Saraceno, Marianthi Markatou, Raktim Mukhopadhyay, Mojgan Golzy"
)
author = "Giovanni Saraceno, Marianthi Markatou, Raktim Mukhopadhyay, Mojgan Golzy"
release = "1.1.1"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.todo",
    "sphinx.ext.mathjax",
    "sphinx.ext.ifconfig",
    "sphinx.ext.githubpages",
    "nbsphinx",
    "sphinx.ext.intersphinx",
    "myst_parser",
]

templates_path = ["_templates"]
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_book_theme"
html_static_path = ["_static"]
html_css_files = ["css/custom.css"]
strip_signature_backslash = True

# -- Options for Latex output -------------------------------------------------
latex_elements = {
    "extraclassoptions": "openany,oneside",
    "preamble": r"""
    \usepackage{fancyhdr}
    \makeatletter
    \fancypagestyle{normal}{
        \fancyhf{}
        \fancyfoot[R]{\py@HeaderFamily\thepage}
        \renewcommand{\headrulewidth}{0pt}
        \renewcommand{\footrulewidth}{0.4pt}
    }
    \makeatother
    """,
}
latex_documents = [
    (
        "index",
        "QuadratiK.tex",
        "QuadratiK",
        author.replace("Raktim", "\\and Raktim"),
        "manual",
    )
]



"""
Contains utility functions for PKBC and PKBD.
"""

import numpy as np


def root_func(x, num_data, alpha_current_h, num_var, mu_denom_h, sum_h_weight_mat):
    """
    Function used for determing the root, used in PKBC.
    """
    return (
        (-2 * num_data * x * alpha_current_h) / (1 - x**2 + 10 ** (-15))
        + num_var * mu_denom_h
        - num_var * x * sum_h_weight_mat
    )


def c_d_lambda(beta, p, lamda):
    """
    Function used for determining root, used in PKBD.
    """
    return (
        -4 * (p - 1) * beta**3
        + (4 * p - lamda**2 * (p - 2) ** 2) * beta**2
        + 2 * p * (p - 2) * lamda**2 * beta
        - p**2 * lamda**2
    )


def calculate_wcss_euclidean(k, memb_best, dat, mu_best):
    """
    Used for computing the euclidean WCSS for a single cluster.
    """
    idx = np.where(memb_best == k)[0]
    return np.sum(np.linalg.norm(dat[idx] - mu_best[k], axis=1) ** 2)


def calculate_wcss_cosine(k, memb_best, dat, mu_best):
    """
    Used for computing the cosine WCSS for a single cluster.
    """
    idx = np.where(memb_best == k)[0]
    return np.sum(dat[idx] * mu_best[k])



import pandas as pd
import numpy as np
from importlib import resources


def load_wireless_data(desc=False, return_X_y=False, as_dataframe=True, scaled=False):
    """
    The wireless data frame has 2000 rows and 8 columns. The first 7 variables
    report the measurements of the Wi-Fi signal strength received from 7 Wi-Fi routers in an
    office location in Pittsburgh (USA). The last column indicates the class labels.

    The function load_wireless_data loads a wireless localization dataset.

    Read more in the :ref:`User Guide <datasets>`.

    Parameters
    ----------

        desc : boolean, optional
            If set to `True`, the function will return the description along with the data.
            If set to `False`, the description will not be included. Defaults to False.

        return_X_y : boolean, optional
            Determines whether the function should return the data as separate arrays (`X` and `y`).
            Defaults to False.

        as_dataframe : boolean, optional
            Determines whether the function should return the data as a pandas DataFrame (Trues)
            or as a numpy array (False). Defaults to True.

        scaled : boolean, optional
            Determines whether or not the data should be scaled. If set to True, the data will be
            divided by its Euclidean norm along each row. Defaults to False.

    Returns
    -------

        (data, target) : tuple, if return_X_y is True
            A tuple of two ndarray. The first containing a 2D array of shape
            (n_samples, n_features) with each row representing one sample and
            each column representing the features. The second ndarray of shape
            (n_samples,) containing the target samples.

        data : pandas.DataFrame, if as_dataframe is True
            Dataframe of the data with shape (n_samples, n_features + class).

        (desc, data, target) : tuple, if desc is True and return_X_y is True
            A tuple of description and two numpy.ndarray. The first containing a 2D
            array of shape (n_samples, n_features) with each row representing
            one sample and each column representing the features. The second
            ndarray of shape (n_samples,) containing the target samples.

        (desc, data) : tuple, if desc is True and as_dataframe is True
            A tuple of description and pandas.DataFrame.
            Dataframe of the data with shape (n_samples, n_features + class)

    References
    ----------
        Rohra, J.G., Perumal, B., Narayanan, S.J., Thakur, P., Bhatt, R.B. (2017).
        User Localization in an Indoor Environment Using Fuzzy Hybrid of Particle Swarm Optimization
        & Gravitational Search Algorithm with Neural Networks. In: Deep, K., et al. Proceedings of
        Sixth International Conference on Soft Computing for Problem Solving. Advances in Intelligent
        Systems and Computing, vol 546. Springer, Singapore. https://doi.org/10.1007/978-981-10-3322-3_27.

    Source
    -------
        Bhatt,Rajen. (2017). Wireless Indoor Localization. UCI Machine Learning Repository.
        https://doi.org/10.24432/C51880.

    Examples
    --------
    >>> from QuadratiK.datasets import load_wireless_data
    >>> X, y = load_wireless_data(return_X_y=True)
    """

    data = np.loadtxt(
        str(
            resources.files("QuadratiK.datasets").joinpath("data/wifi_localization.txt")
        )
    )

    if scaled:
        data[:, :-1] = data[:, :-1] / np.linalg.norm(
            data[:, :-1], axis=1, keepdims=True
        )

    feature_names = ["WS1", "WS2", "WS3", "WS4", "WS5", "WS6", "WS7", "Class"]

    if desc:
        desc_file = resources.files("QuadratiK.datasets").joinpath(
            "data/wireless_localization_dataset.rst"
        )
        fdescr = desc_file.read_text()

    if return_X_y:
        X = data[:, :-1]
        y = data[:, -1].astype(int)
        if desc:
            return (fdescr, X, y)
        else:
            return (X, y)

    if as_dataframe:
        data = pd.DataFrame(data, columns=feature_names)
        data["Class"] = data["Class"].astype(int)
        if desc:
            return (fdescr, data)
        else:
            return data
    else:
        return data


def load_wisconsin_breast_cancer_data(
    desc=False, return_X_y=False, as_dataframe=True, scaled=False
):
    """
    The Wisconsin breast cancer dataset data frame has 569 rows and 31 columns. The first 30 variables
    report the features are computed from a digitized image of a fine needle aspirate (FNA) of a breast mass.
    They describe characteristics of the cell nuclei present in the image.
    The last column indicates the class labels (Benign = 0 or Malignant = 1).

    The function load_wisconsin_breast_cancer_data loads the Breast Cancer Wisconsin (Diagnostic).

    Read more in the :ref:`User Guide <datasets>`.

    Parameters
    ----------

        desc : boolean, optional
            If set to `True`, the function will return the description along with the data.
            If set to `False`, the description will not be included. Defaults to False.

        return_X_y : boolean, optional
            Determines whether the function should return the data as separate arrays (`X` and `y`).
            Defaults to False.

        as_dataframe : boolean, optional
            Determines whether the function should return the data as a pandas DataFrame (Trues)
            or as a numpy array (False). Defaults to True.

        scaled : boolean, optional
            Determines whether or not the data should be scaled. If set to True, the data will be
            divided by its Euclidean norm along each row. Defaults to False.

    Returns
    -------

        (data, target) : tuple, if return_X_y is True
            A tuple of two ndarray. The first containing a 2D array of shape
            (n_samples, n_features) with each row representing one sample and
            each column representing the features. The second ndarray of shape
            (n_samples,) containing the target samples.

        data : pandas.DataFrame, if as_dataframe is True
            Dataframe of the data with shape (n_samples, n_features + class).

        (desc, data, target) : tuple, if desc is True and return_X_y is True
            A tuple of description and two numpy.ndarray. The first containing a 2D
            array of shape (n_samples, n_features) with each row representing
            one sample and each column representing the features. The second
            ndarray of shape (n_samples,) containing the target samples.

        (desc, data) : tuple, if desc is True and as_dataframe is True
            A tuple of description and pandas.DataFrame.
            Dataframe of the data with shape (n_samples, n_features + class)

    References
    ----------
        Street, W. N., Wolberg, W. H., & Mangasarian, O. L. (1993, July).
        Nuclear feature extraction for breast tumor diagnosis.
        In Biomedical image processing and biomedical visualization (Vol. 1905, pp. 861-870). SPIE.

    Source
    -------
        Wolberg,William, Mangasarian,Olvi, Street,Nick, and Street,W.. (1995).
        Breast Cancer Wisconsin (Diagnostic). UCI Machine Learning Repository. https://doi.org/10.24432/C5DW2B.

    Examples
    --------
    >>> from QuadratiK.datasets import load_wisconsin_breast_cancer_data
    >>> X, y = load_wisconsin_breast_cancer_data(return_X_y=True)
    """

    data = np.loadtxt(
        str(
            resources.files("QuadratiK.datasets").joinpath(
                "data/wisconsin_breast_cancer.txt"
            )
        )
    )

    if scaled:
        data[:, :-1] = data[:, :-1] / np.linalg.norm(
            data[:, :-1], axis=1, keepdims=True
        )

    feature_names = [
        "radius1",
        "texture1",
        "perimeter1",
        "area1",
        "smoothness1",
        "compactness1",
        "concavity1",
        "concave_points1",
        "symmetry1",
        "fractal_dimension1",
        "radius2",
        "texture2",
        "perimeter2",
        "area2",
        "smoothness2",
        "compactness2",
        "concavity2",
        "concave_points2",
        "symmetry2",
        "fractal_dimension2",
        "radius3",
        "texture3",
        "perimeter3",
        "area3",
        "smoothness3",
        "compactness3",
        "concavity3",
        "concave_points3",
        "symmetry3",
        "fractal_dimension3",
        "Class",
    ]

    if desc:
        desc_file = resources.files("QuadratiK.datasets").joinpath(
            "data/wisconsin_breast_cancer_dataset.rst"
        )
        fdescr = desc_file.read_text()

    if return_X_y:
        X = data[:, :-1]
        y = data[:, -1].astype(int)
        if desc:
            return (fdescr, X, y)
        else:
            return (X, y)

    if as_dataframe:
        data = pd.DataFrame(data, columns=feature_names)
        data["Class"] = data["Class"].astype(int)
        if desc:
            return (fdescr, data)
        else:
            return data
    else:
        return data


def load_wine_data(desc=False, return_X_y=False, as_dataframe=True, scaled=False):
    """
    The wine data frame has 178 rows and 14 columns. The first 13 variables
    report 13 constituents found in each of the three types of wines.
    The last column indicates the class labels (1,2 or 3).

    The function load_wine_data loads the Wine dataset.

    Read more in the :ref:`User Guide <datasets>`.

    Parameters
    ----------

        desc : boolean, optional
            If set to `True`, the function will return the description along with the data.
            If set to `False`, the description will not be included. Defaults to False.

        return_X_y : boolean, optional
            Determines whether the function should return the data as separate arrays (`X` and `y`).
            Defaults to False.

        as_dataframe : boolean, optional
            Determines whether the function should return the data as a pandas DataFrame (Trues)
            or as a numpy array (False). Defaults to True.

        scaled : boolean, optional
            Determines whether or not the data should be scaled. If set to True, the data will be
            divided by its Euclidean norm along each row. Defaults to False.

    Returns
    -------

        (data, target) : tuple, if return_X_y is True
            A tuple of two ndarray. The first containing a 2D array of shape
            (n_samples, n_features) with each row representing one sample and
            each column representing the features. The second ndarray of shape
            (n_samples,) containing the target samples.

        data : pandas.DataFrame, if as_dataframe is True
            Dataframe of the data with shape (n_samples, n_features + class).

        (desc, data, target) : tuple, if desc is True and return_X_y is True
            A tuple of description and two numpy.ndarray. The first containing a 2D
            array of shape (n_samples, n_features) with each row representing
            one sample and each column representing the features. The second
            ndarray of shape (n_samples,) containing the target samples.

        (desc, data) : tuple, if desc is True and as_dataframe is True
            A tuple of description and pandas.DataFrame.
            Dataframe of the data with shape (n_samples, n_features + class)

    References
    ----------
        Aeberhard, S., Coomans, D., & De Vel, O. (1994). Comparative analysis of statistical pattern recognition
        methods in high dimensional settings. Pattern Recognition, 27(8), 1065-1077.

    Source
    -------
        Aeberhard,Stefan and Forina,M.. (1991). Wine.
        UCI Machine Learning Repository. https://doi.org/10.24432/C5PC7J.

    Examples
    --------
    >>> from QuadratiK.datasets import load_wine_data
    >>> X, y = load_wine_data(return_X_y=True)
    """

    data = np.loadtxt(
        str(resources.files("QuadratiK.datasets").joinpath("data/wine.txt"))
    )

    if scaled:
        data[:, :-1] = data[:, :-1] / np.linalg.norm(
            data[:, :-1], axis=1, keepdims=True
        )

    feature_names = [
        "Alcohol",
        "Malicacid",
        "Ash",
        "Alcalinity_of_ash",
        "Magnesium",
        "Total_phenols",
        "Flavanoids",
        "Nonflavanoid_phenols",
        "Proanthocyanins",
        "Color_intensity",
        "Hue",
        "0D280_0D315_of_diluted_wines",
        "Proline",
        "Class",
    ]

    if desc:
        desc_file = resources.files("QuadratiK.datasets").joinpath(
            "data/wine_dataset.rst"
        )
        fdescr = desc_file.read_text()

    if return_X_y:
        X = data[:, :-1]
        y = data[:, -1].astype(int)
        if desc:
            return (fdescr, X, y)
        else:
            return (X, y)

    if as_dataframe:
        data = pd.DataFrame(data, columns=feature_names)
        data["Class"] = data["Class"].astype(int)
        if desc:
            return (fdescr, data)
        else:
            return data
    else:
        return data



"""
Contains the PKBC functionality of the UI
"""

import importlib
import copy
from io import BytesIO
import streamlit as st
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import plotly.graph_objects as go
from plotly.subplots import make_subplots
from sklearn.preprocessing import LabelEncoder
from sklearn.utils.parallel import Parallel, delayed


kt = importlib.import_module("QuadratiK.kernel_test").KernelTest
qq_plot = importlib.import_module("QuadratiK.tools").qq_plot
stats = importlib.import_module("QuadratiK.tools").stats

pkbc = importlib.import_module("QuadratiK.spherical_clustering").PKBC
sphere3d = importlib.import_module("QuadratiK.tools").sphere3d
circle2d = importlib.import_module("QuadratiK.tools").plot_clusters_2d
utils = importlib.import_module("QuadratiK.tools")._utils


st.title("Poisson Kernel based Clustering")
st.write(
    "Performs the Poisson kernel-based clustering algorithm on the Sphere based on \
    the Poisson kernel-based densities"
)

with st.expander("Click to view code"):
    code_python = """
    # In case you do not have the true labels, do not read y.
    X,y = Read the data and the cluster label files. 

    from QuadratiK.spherical_clustering import PKBC
    cluster_fit = PKBC(num_clust = Input the number of clusters).fit(X)
    """
    st.code(code_python, language="python")

    code_R = """
    library(QuadratiK)
    data <- Read Data here
    res_pk <- pkbc(as.matrix(data),2:10)
    labels <- Read labels here
    res_validation <- validation(res_pk, true_label = labels)
    """
    st.code(code_R, language="r")

head = st.checkbox("**Select, if the header is present in the data file.**", value=True)
delim = st.text_input("**Enter the delimiter**", ",")
data = st.file_uploader(
    "**Please Upload the data file**",
    accept_multiple_files=False,
    type=[".csv", ".txt"],
)

if data is not None:
    st.success(data.name + " Uploaded Successfully!")
    if head:
        data = pd.read_csv(data, sep=delim)
    else:
        data = pd.read_csv(data, sep=delim, header=None)
    data.columns = range(data.shape[1])
    y_true_available = st.checkbox(
        "**Select the checkbox if true labels for the data points are available**",
        value=False,
    )

    col_number = None
    if y_true_available:
        try:
            col_number = int(
                st.number_input(
                    "**Enter the column in the datafile that contains the label (start from 0)**",
                    value=None,
                    step=1,
                )
            )
        except:
            st.error("Enter a valid column number")

    else:
        pass
    num_clusters = int(
        st.number_input("**Enter the number of clusters**", value=2, step=1)
    )

if data is not None:
    if y_true_available:
        if col_number is not None:
            try:
                x = copy.copy(data.drop(columns=[col_number]))
                x = np.array(x, dtype=np.float64)
                y = data[col_number]
                le = LabelEncoder()
                le.fit(y)
                y = le.transform(y)
            except:
                st.error("Please check the input data file")
    else:
        x = copy.copy(data)

    try:
        with st.spinner("getting results ready..."):
            cluster_fit = pkbc(num_clust=num_clusters).fit(x)
            y_pred = cluster_fit.labels_

            validation_metrics, elbow_plot = cluster_fit.validation(y)

            st.dataframe(validation_metrics)

            st.write("**Results:**")
            clustering_results_json = {}
            clustering_results_json["post_probs"] = cluster_fit.post_probs_
            clustering_results_json["loglik"] = cluster_fit.loglik_
            clustering_results_json["mu"] = cluster_fit.mu_
            clustering_results_json["alpha"] = cluster_fit.alpha_
            clustering_results_json["rho"] = cluster_fit.rho_
            clustering_results_json["final_membership"] = cluster_fit.labels_
            clustering_results_json["euclidean_wcss"] = cluster_fit.euclidean_wcss_
            clustering_results_json["cosine_wcss"] = cluster_fit.euclidean_wcss_
            clustering_results_json["log_lik_vec"] = cluster_fit.log_lik_vec
            clustering_results_json["num_iter_per_run"] = cluster_fit.num_iter_per_run
            clustering_results_json["num_clust"] = cluster_fit.num_clust
            clustering_results_json["max_iter"] = cluster_fit.max_iter
            clustering_results_json["stopping_rule"] = cluster_fit.stopping_rule
            clustering_results_json["init_method"] = cluster_fit.init_method
            clustering_results_json["num_init"] = cluster_fit.num_init

            st.json(clustering_results_json, expanded=False)
            st.download_button(
                "Click to Download the clustering results",
                str(clustering_results_json),
                "Clustering Results.txt",
                "txt",
                key="download-cluster-res",
            )

            st.success("Done!")
    except:
        st.error("Please check the input data file")

st.header("K-Sample Test for the identified clusters", divider="grey")

num_iter = st.number_input(
    "Enter a value of number of iterations to be used for critical value estimation",
    value=500,
)
h = st.number_input("Enter a value of tuning parameter h", value=1)
b = st.number_input(
    "Enter a value for the proportion of subsampling samples to be used", value=0.9
)

if data is not None:
    num_iter = int(num_iter)
    h_val = float(h)
    b = float(b)

    try:
        x_copy = x / np.linalg.norm(x, axis=1, keepdims=True)
        y_pred = cluster_fit.labels_[num_clusters]

        with st.spinner("getting results ready..."):
            k_samp_test = kt(
                h=h_val, num_iter=num_iter, b=b, centering_type="nonparam"
            ).test(x=x_copy, y=y_pred)

            st.write("**Results:**")

            res = pd.DataFrame()
            res["Value"] = [
                k_samp_test.test_type_,
                k_samp_test.dn_test_statistic_,
                k_samp_test.dn_cv_,
                k_samp_test.dn_h0_rejected_,
                k_samp_test.var_dn_,
                k_samp_test.trace_test_statistic_,
                k_samp_test.trace_cv_,
                k_samp_test.trace_h0_rejected_,
                k_samp_test.var_trace_,
            ]
            res = res.set_axis(
                [
                    "Test Type",
                    "Dn Test Statistic",
                    "Dn Critical Value",
                    "Dn Reject H0",
                    "Var Dn",
                    "Trace Test Statistic",
                    "Trace Critical Value",
                    "Trace Reject H0",
                    "Var Trace",
                ]
            )

            st.dataframe(res, width=800)
            csv_res = res.to_csv().encode()
            st.download_button(
                "Click to Download the test results",
                csv_res,
                "K_Sample_Test_results.csv",
                "text/csv",
                key="download-test-res",
            )

            summary_stats_df = k_samp_test.stats()
            st.dataframe(summary_stats_df)

            csv_stats = summary_stats_df.to_csv(index=True).encode()
            st.download_button(
                "Click to Download the summary statistics",
                csv_stats,
                "Statistics.csv",
                "text/csv",
                key="download-summary",
            )
    except:
        st.error("Check data file and selected parameters.")

st.header("Visualizations", divider="grey")

st.subheader("Elbow Plot")

with st.expander("Click to view code"):
    elbow_code = """
    import matplotlib.pyplot as plt
    wcss_list = []
    for clus in range(2,10):
        cluster_fit = PKBC(num_clust=clus).fit(X)
        wcss_list.append(cluster_fit.euclidean_wcss_)
        
    plt.plot(list(range(2,10)),wcss_list, "--o")
    plt.xlabel("Number of Cluster")
    plt.ylabel("Within Cluster Sum of Squares (WCSS)")
    plt.title("Elbow Plot")
    """
    st.code(elbow_code, language="python")


def get_wcss_euclid(x, k):
    cluster_fitting = pkbc(num_clust=k).fit(x)
    return cluster_fitting.euclidean_wcss_[k]


def get_wcss_cosine(x, k):
    cluster_fitting = pkbc(num_clust=k).fit(x)
    return cluster_fitting.cosine_wcss_[k]


if data is not None:
    elbow_clusters = int(
        st.number_input(
            "Enter the total number of clusters for which elbow plot to be shown",
            value=10,
            step=1,
        )
    )
    with st.spinner("Generating the elbow plot ..."):
        try:
            wcss_list_euclid = Parallel(n_jobs=4)(
                delayed(get_wcss_euclid)(x, k) for k in range(2, elbow_clusters)
            )

            wcss_list_cosine = Parallel(n_jobs=4)(
                delayed(get_wcss_cosine)(x, k) for k in range(2, elbow_clusters)
            )
            fig, axs = plt.subplots(1, 2, figsize=(15, 6))
            axs[0].plot(list(range(2, elbow_clusters)), wcss_list_euclid, "--o")
            axs[0].set_xlabel("Number of Cluster")
            axs[0].set_ylabel("Euclidean Within Cluster Sum of Squares (WCSS)")
            axs[0].set_title("Elbow Plot WCSS Euclidean")

            axs[1].plot(list(range(2, elbow_clusters)), wcss_list_cosine, "--o")
            axs[1].set_xlabel("Number of Cluster")
            axs[1].set_ylabel(
                "Euclidean Within Cluster Sum of Squares (Cosine Similarity)"
            )
            axs[1].set_title("Elbow Plot WCSS Cosine Similarity")
            st.pyplot(fig)
        except:
            st.error("Please ensure that the data file is loaded")


st.subheader("Data on Sphere")
with st.expander("Click to view code"):
    viz_code = """
    from QuadratiK.tools import sphere3d
    sphere3d(X,y)

    # or in case the input data is 2d

    from QuadratiK.tools import plot_clusters_2d
    plot_clusters_2d(X,y)
    """
    st.code(viz_code, language="python")

if data is not None:
    try:
        if x.shape[1] > 2:
            with st.spinner("Plotting the data points on sphere ..."):
                r = 1
                pi = np.pi
                cos = np.cos
                sin = np.sin
                phi, theta = np.mgrid[0.0:pi:100j, 0.0 : 2.0 * pi : 100j]
                x1 = r * sin(phi) * cos(theta)
                y1 = r * sin(phi) * sin(theta)
                z1 = r * cos(phi)
                xx, yy, zz = utils._extract_3d(x)

                fig = make_subplots(
                    rows=1,
                    cols=2,
                    specs=[[{"type": "scatter3d"}, {"type": "scatter3d"}]],
                    subplot_titles=(
                        "Colored by Predicted Class",
                        "Colored by True Class",
                    ),
                )
                fig.append_trace(
                    go.Surface(
                        x=x1,
                        y=y1,
                        z=z1,
                        colorscale=[[0, "#DCDCDC"], [1, "#DCDCDC"]],
                        opacity=0.5,
                        showscale=False,
                    ),
                    row=1,
                    col=1,
                )
                fig.append_trace(
                    go.Scatter3d(
                        x=xx,
                        y=yy,
                        z=zz,
                        mode="markers",
                        marker=dict(
                            size=5, color=y_pred, colorscale="turbo", showscale=False
                        ),
                    ),
                    row=1,
                    col=1,
                )

                if y_true_available:
                    fig.append_trace(
                        go.Surface(
                            x=x1,
                            y=y1,
                            z=z1,
                            colorscale=[[0, "#DCDCDC"], [1, "#DCDCDC"]],
                            opacity=0.5,
                            showscale=False,
                        ),
                        row=1,
                        col=2,
                    )
                    fig.append_trace(
                        go.Scatter3d(
                            x=xx,
                            y=yy,
                            z=zz,
                            mode="markers",
                            marker=dict(
                                size=5, color=y, colorscale="cividis", showscale=False
                            ),
                        ),
                        row=1,
                        col=2,
                    )

                fig.update_layout(
                    title="",
                    scene=dict(
                        xaxis=dict(range=[-1, 1]),
                        yaxis=dict(range=[-1, 1]),
                        zaxis=dict(range=[-1, 1]),
                        aspectmode="data",
                    ),
                )
                fig.update_layout(showlegend=False)
                st.plotly_chart(fig, use_container_width=True)
        else:
            fig1 = plt.figure(figsize=(8, 6))
            fig1 = circle2d(x, y_pred)
            plt.title("Colored by Predicted Class")
            st.pyplot(fig1)

            if y_true_available:
                fig2 = plt.figure(figsize=(8, 6))
                fig2 = circle2d(x, y)
                plt.title("Colored by True Class")
                st.pyplot(fig2)
    except:
        st.error("Please ensure that the data file is loaded")

st.markdown(
    r"""
    <style>
        .reportview-container {
            margin-top: -2em;
        }
        #MainMenu {visibility: hidden;}
        .stDeployButton {display:none;}
        footer {visibility: hidden;}
        #stDecoration {display:none;}
    </style>
""",
    unsafe_allow_html=True,
)



"""
Poisson kernel-based quadratic distance test of Uniformity on the Sphere
"""

import importlib
import numpy as np
import pandas as pd
from scipy.stats import chi2
from tabulate import tabulate

from ._utils import dof, stat_poisson_unif
from ._cv_functions import poisson_cv


time_decorator = importlib.import_module(
    "QuadratiK.tools._utils"
).class_method_call_timing
stats = importlib.import_module("QuadratiK.tools").stats


class PoissonKernelTest:
    """
    Class for Poisson kernel-based quadratic distance tests
    of Uniformity on the Sphere.

    Parameters
    ----------
        rho : float
            The value of concentration parameter used for the
            Poisson kernel function.

        num_iter : int, optional
            Number of iterations for critical value estimation of U-statistic.

        quantile : float, optional
            The quantile to use for critical value estimation.

        random_state : int, None, optional.
            Seed for random number generation. Defaults to None.

        n_jobs : int, optional.
            n_jobs specifies the maximum number of concurrently running workers.
            If 1 is given, no joblib parallelism is used at all, which is useful for debugging.
            For more information on joblib n_jobs refer
            to - https://joblib.readthedocs.io/en/latest/generated/joblib.Parallel.html.
            Defaults to 8.

    Attributes
    -----------
        test_type\\_ : str
            The type of test performed on the data.

        execution_time : float
            Time taken for the test method to execute.

        u_statistic_h0\\_ : boolean
            A logical value indicating whether or not the null hypothesis
            is rejected according to Un.

        u_statistic_un\\_ : float
            The value of the U-statistic.

        u_statistic_cv\\_ : float
            The empirical critical value for Un.

        v_statistic_h0\\_ : boolean
            A logical value indicating whether or not the null hypothesis is
            rejected according to Vn.

        v_statistic_vn\\_ : float
            The value of the V-statistic.

        v_statistic_cv\\_ : float
            The critical value for Vn computed following the asymptotic distribution.

    References
    -----------
        Ding Y., Markatou M., Saraceno G. (2023). “Poisson Kernel-Based Tests for
        Uniformity on the d-Dimensional Sphere.” Statistica Sinica. doi: doi:10.5705/ss.202022.0347.

    Examples
    ---------
    >>> import numpy as np
    >>> np.random.seed(0)
    >>> from QuadratiK.poisson_kernel_test import PoissonKernelTest
    >>> # data generation
    >>> z = np.random.normal(size=(200, 3))
    >>> data_unif = z / np.sqrt(np.sum(z**2, axis=1, keepdims=True))
    >>> #performing the uniformity test
    >>> unif_test = PoissonKernelTest(rho = 0.7, random_state=42).test(data_unif)
    >>> print(unif_test)
    ... PoissonKernelTest(
        Test Type=Poisson Kernel-based quadratic distance test of Uniformity on the Sphere,
        Execution Time=3.3801472187042236 seconds,
        U-Statistic=0.5977824645431915,
        U-Statistic Critical Value=1.6128083124315886,
        U-Statistic Null Hypothesis Rejected=False,
        V-Statistic=19.722614852087553,
        V-Statistic Critical Value=23.229486935225513,
        V-Statistic Null Hypothesis Rejected=False,
        Selected concentration parameter rho=0.7,
        )
    """

    def __init__(self, rho, num_iter=300, quantile=0.95, random_state=None, n_jobs=8):
        self.rho = rho
        self.num_iter = num_iter
        self.quantile = quantile
        self.random_state = random_state
        self.n_jobs = n_jobs

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(\n"
            f"  Test Type={self.test_type_},\n"
            f"  Execution Time={self.execution_time} seconds,\n"
            f"  U-Statistic={self.u_statistic_un_},\n"
            f"  U-Statistic Critical Value={self.u_statistic_cv_},\n"
            f"  U-Statistic Null Hypothesis Rejected={self.u_statistic_h0_},\n"
            f"  V-Statistic={self.v_statistic_vn_},\n"
            f"  V-Statistic Critical Value={self.v_statistic_cv_},\n"
            f"  V-Statistic Null Hypothesis Rejected={self.v_statistic_h0_},\n"
            f"  Selected concentration parameter rho={self.rho},\n"
            f")"
        )

    @time_decorator
    def test(self, x):
        """
        Performs the Poisson kernel-based quadratic distance Goodness-of-fit tests for
        Uniformity for spherical data using the Poisson kernel with concentration parameter :math:`rho`.

        Parameters
        ----------
            x : numpy.ndarray, pandas.DataFrame
                a numeric d-dim matrix of data points on the Sphere :math:`S^{(d-1)}`.

        Returns
        -------
            self : object
                Fitted estimator.
        """
        self.x = x
        if isinstance(x, np.ndarray):
            if self.x.ndim == 1:
                self.x = self.x.reshape(-1, 1)
        elif isinstance(self.x, pd.DataFrame):
            self.x = self.x.to_numpy()
        elif not isinstance(self.x, (np.ndarray, pd.DataFrame)):
            raise TypeError("x must be a numpy array or a pandas dataframe")

        if (self.quantile <= 0) or (self.quantile > 1):
            raise ValueError(
                "Quantile indicates the level used for the critical \
                    value computation. It must be in (0,1]."
            )

        if not (isinstance(self.rho, (int, float)) and (0 < self.rho <= 1)):
            raise ValueError(
                "rho indicates the concentration parameter \
                    of the Poisson kernel, it must be in (0,1)."
            )

        if not isinstance(self.random_state, (int, type(None))):
            raise ValueError("Please specify a integer or None random_state")

        method = (
            "Poisson Kernel-based quadratic distance test of Uniformity on the Sphere"
        )

        n, d = self.x.shape
        pk = stat_poisson_unif(self.x, self.rho)
        var_un = (2 / (n * (n - 1))) * (
            (1 + np.power(self.rho, 2)) / np.power(1 - np.power(self.rho, 2), (d - 1))
            - 1
        )
        dof_val = dof(d, self.rho)
        qu_q = chi2.ppf(0.95, dof_val["DOF"])
        cv_vn = dof_val["Coefficient"] * qu_q
        cv_un = poisson_cv(
            d=d,
            size=n,
            rho=self.rho,
            num_iter=self.num_iter,
            quantile=self.quantile,
            random_state=self.random_state,
            n_jobs=self.n_jobs,
        ) / np.sqrt(var_un)

        self.test_type_ = method
        self.u_statistic_h0_ = (pk[0] / np.sqrt(var_un)) > cv_un
        self.u_statistic_un_ = pk[0] / np.sqrt(var_un)
        self.u_statistic_cv_ = cv_un

        self.v_statistic_h0_ = pk[1] > cv_vn
        self.v_statistic_vn_ = pk[1]
        self.v_statistic_cv_ = cv_vn

        return self

    def stats(self):
        """
        Function to generate descriptive statistics.

        Returns
        -------
            summary_stats_df : pandas.DataFrame
                Dataframe of descriptive statistics.
        """
        summary_stats_df = stats(self.x)
        return summary_stats_df.round(4)

    def summary(self, print_fmt="simple_grid"):
        """
        Summary function generates a table for
        the Poisson kernel test results and the summary statistics.

        Parameters
        ----------
            print_fmt : str, optional.
                Used for printing the output in the desired format.
                Supports all available options in tabulate,
                see here: https://pypi.org/project/tabulate/.
                Defaults to "simple_grid".

        Returns
        --------
            summary : str
                A string formatted in the desired output
                format with the kernel test results and summary statistics.
        """
        index_labels = [
            "Test Statistic",
            "Critical Value",
            "H0 is rejected (1 = True, 0 = False)",
        ]
        test_summary = {
            "U-Statistic": [
                self.u_statistic_un_,
                self.u_statistic_cv_,
                self.u_statistic_h0_,
            ],
            "V-Statistic": [
                self.v_statistic_vn_,
                self.v_statistic_cv_,
                self.v_statistic_h0_,
            ],
        }
        res = pd.DataFrame(test_summary, index=index_labels)

        summary_stats_df = self.stats().round(4)

        if print_fmt == "html":
            summary_string = (
                "Time taken for execution: {} seconds".format(self.execution_time)
                + "<br>Test Results <br>"
                + f"<br>{self.test_type_} <br>"
                + tabulate(res, tablefmt=print_fmt, headers=res.columns)
                + f"<br>Concentration parameter rho: {self.rho}<br>"
                + "<br> Summary Statistics <br>"
                + tabulate(
                    summary_stats_df,
                    tablefmt=print_fmt,
                    headers=summary_stats_df.columns,
                )
                + "<br>"
            )
        else:
            summary_string = (
                "Time taken for execution: {:.3f} seconds".format(self.execution_time)
                + "\nTest Results \n"
                + f"{self.test_type_} \n"
                + tabulate(res, tablefmt=print_fmt, headers=res.columns)
                + f"\nConcentration parameter rho: {self.rho}\n"
                + "\nSummary Statistics \n"
                + tabulate(
                    summary_stats_df,
                    tablefmt=print_fmt,
                    headers=summary_stats_df.columns,
                )
            )

        return summary_string



"""
Kernel-based quadratic distance Goodness-of-Fit tests
"""

import importlib
import numpy as np
import pandas as pd
from tabulate import tabulate
from scipy.stats import chi2

from ._utils import (
    stat_normality_test,
    stat_two_sample,
    stat_ksample,
    dof_normality_test,
    variance_normality_test,
)
from ._cv_functions import cv_twosample, cv_normality, cv_ksample
from ._h_selection import select_h

time_decorator = importlib.import_module(
    "QuadratiK.tools._utils"
).class_method_call_timing
stats = importlib.import_module("QuadratiK.tools").stats


class KernelTest:
    """
    Class for performing the kernel-based quadratic distance goodness-of-fit tests using
    the Gaussian kernel with tuning parameter h. Depending on the input `y` the function performs
    the test of multivariate normality, the non-parametric two-sample tests or the k-sample tests.

    Parameters
    ----------
        h : float, optional
            Bandwidth for the kernel function.

        method : str, optional
            The method used for critical value estimation ("subsampling", "bootstrap",
            or "permutation").

        num_iter : int, optional
            The number of iterations to use for critical value estimation. Defaults to 150.

        b : float, optional
            The size of the subsamples used in the subsampling algorithm. Defaults to 0.9 i.e.
            `0.9N` samples are used, where `N` represents the total sample size.

        quantile : float, optional
            The quantile to use for critical value estimation. Defaults to 0.95.

        mu_hat : numpy.ndarray, optional
            Mean vector for the reference distribution. Defaults to None.

        sigma_hat : numpy.ndarray, optional
            Covariance matrix of the reference distribution. Defaults to None.

        alternative : str, optional
            String indicating the type of alternative to be used for calculating "h"
            by the tuning parameter selection algorithm when h is not provided.
            Defaults to 'None'

        k_threshold : int, optional
            Maximum number of groups allowed. Defaults to 10. Change in case of more than 10 groups.

        random_state : int, None, optional.
            Seed for random number generation. Defaults to None.

        n_jobs : int, optional.
            n_jobs specifies the maximum number of concurrently
            running workers. If 1 is given, no joblib parallelism
            is used at all, which is useful for debugging. For more
            information on joblib n_jobs refer to -
            https://joblib.readthedocs.io/en/latest/generated/joblib.Parallel.html.
            Defaults to 8.

    Attributes
    ----------
        For Normality Test:
            test_type\\_ : str
                The type of test performed on the data.

            execution_time : float
                Time taken for the test method to execute.

            un_h0_rejected\\_ : boolean
                Whether the null hypothesis using Un is rejected (True) or not (False).

            vn_h0_rejected\\_ : boolean
                Whether the null hypothesis using Vn is rejected (True) or not (False).

            un_test_statistic\\_ : float
                Un Test statistic of the perfomed test type.

            vn_test_statistic\\_ : float
                Vn Test statistic of the perfomed test type.

            un_cv\\_ : float
                Critical value for Un.

            un_cv\\_ : float
                Critical value for Vn.

        For Two-Sample and K-Sample Test:
            test_type\\_ : str
                The type of test performed on the data.

            execution_time : float
                Time taken for the test method to execute.

            dn_h0_rejected\\_ : boolean
                Whether the null hypothesis using Un is rejected (True) or not (False).

            dn_test_statistic\\_ : float
                Un Test statistic of the perfomed test type.

            dn_cv\\_ : float
                Critical value for Un.

            trace_h0_rejected\\_ : boolean
                Whether the null hypothesis using trace statistic is rejected (True) or not (False).

            trace_test_statistic\\_ : float
                Trace Test statistic of the perfomed test type.

            trace_cv\\_ : float
                Critical value for trace statistic.

            cv_method\\_ : str
                Critical value method used for performing the test.

    References
    -----------
        Markatou M., Saraceno G., Chen Y (2023). “Two- and k-Sample Tests Based on Quadratic Distances.
        ”Manuscript, (Department of Biostatistics, University at Buffalo).

        Lindsay BG, Markatou M. & Ray S. (2014) Kernels, Degrees of Freedom, and
        Power Properties of Quadratic Distance Goodness-of-Fit Tests, Journal of the American Statistical
        Association, 109:505, 395-410, DOI: 10.1080/01621459.2013.836972.

    Examples
    --------
    >>> import numpy as np
    >>> np.random.seed(78990)
    >>> from QuadratiK.kernel_test import KernelTest
    >>> # data generation
    >>> data_norm = np.random.multivariate_normal(mean = np.zeros(4), cov = np.eye(4),size = 500)
    >>> # performing the normality test
    >>> normality_test = KernelTest(h=0.4, num_iter=150, method= "subsampling", random_state=42).test(data_norm)
    >>> print(normality_test)
    ... KernelTest(
        Test Type=Kernel-based quadratic distance Normality test,
        Execution Time=3.198080062866211 seconds,
        U-Statistic=-0.1145936604628874,
        U-Statistic Critical Value=1.1593122985543514,
        U-Statistic Null Hypothesis Rejected=False,
        U-Statistic Variance=1.108021332522181e-08,
        V-Statistic=0.977955027161687,
        V-Statistic Critical Value=42.460022848761945,
        V-Statistic Null Hypothesis Rejected=False,
        Selected tuning parameter h=0.4
        )

    >>> import numpy as np
    >>> np.random.seed(0)
    >>> from scipy.stats import skewnorm
    >>> from QuadratiK.kernel_test import KernelTest
    >>> # data generation
    >>> X_2 = np.random.multivariate_normal(mean = np.zeros(4), cov = np.eye(4), size=200)
    >>> Y_2 = skewnorm.rvs(size=(200, 4),loc=np.zeros(4), scale=np.ones(4),a=np.repeat(0.5,4), random_state=20)
    >>> # performing the two sample test
    >>> two_sample_test = KernelTest(h = 2,num_iter = 150, random_state=42).test(X_2,Y_2)
    >>> print(two_sample_test)
    ... KernelTest(
        Test Type=Kernel-based quadratic distance two-sample test,
        Execution Time=0.36570000648498535 seconds,
        Dn-Statistic=5.061212999055004,
        Dn-Statistic Critical Value=0.4901155246432661,
        Dn-Statistic Null Hypothesis Rejected=True,
        Dn-Statistic Variance=3.037711857184588e-10,
        Trace-Statistic=15.751718163734266,
        Trace-Statistic Critical Value=1.525782865913332,
        Trace-Statistic Null Hypothesis Rejected=True,
        Trace-Statistic Variance=7.879780877050946e-12,
        Selected tuning parameter h=2,
        Critical Value Method=subsampling
        )
    """

    def __init__(
        self,
        h=None,
        method="subsampling",
        num_iter=150,
        b=0.9,
        quantile=0.95,
        mu_hat=None,
        sigma_hat=None,
        centering_type="nonparam",
        alternative=None,
        k_threshold=10,
        random_state=None,
        n_jobs=8,
    ):
        self.h = h
        self.method = method
        self.num_iter = num_iter
        self.b = b
        self.quantile = quantile
        self.mu_hat = mu_hat
        self.sigma_hat = sigma_hat
        self.centering_type = centering_type
        self.alternative = alternative
        self.k_threshold = k_threshold
        self.n_jobs = n_jobs
        self.random_state = random_state

    def __repr__(self):
        if self.vn_test_statistic_ is not None:
            return (
                f"{self.__class__.__name__}(\n"
                f"  Test Type={self.test_type_},\n"
                f"  Execution Time={self.execution_time} seconds,\n"
                f"  U-Statistic={self.un_test_statistic_},\n"
                f"  U-Statistic Critical Value={self.un_cv_},\n"
                f"  U-Statistic Null Hypothesis Rejected={self.un_h0_rejected_},\n"
                f"  U-Statistic Variance={self.var_un_},\n"
                f"  V-Statistic={self.vn_test_statistic_},\n"
                f"  V-Statistic Critical Value={self.vn_cv_},\n"
                f"  V-Statistic Null Hypothesis Rejected={self.vn_h0_rejected_},\n"
                f"  Selected tuning parameter h={self.h}\n"
                f")"
            )
        else:
            return (
                f"{self.__class__.__name__}(\n"
                f"  Test Type={self.test_type_},\n"
                f"  Execution Time={self.execution_time} seconds,\n"
                f"  Dn-Statistic={self.dn_test_statistic_},\n"
                f"  Dn-Statistic Critical Value={self.dn_cv_},\n"
                f"  Dn-Statistic Null Hypothesis Rejected={self.dn_h0_rejected_},\n"
                f"  Dn-Statistic Variance={self.var_dn_},\n"
                f"  Trace-Statistic={self.trace_test_statistic_},\n"
                f"  Trace-Statistic Critical Value={self.trace_cv_},\n"
                f"  Trace-Statistic Null Hypothesis Rejected={self.trace_h0_rejected_},\n"
                f"  Trace-Statistic Variance={self.var_trace_},\n"
                f"  Selected tuning parameter h={self.h},\n"
                f"  Critical Value Method={self.cv_method_}\n"
                f")"
            )

    @time_decorator
    def test(self, x, y=None):
        """
        Function to perform the kernel-based quadratic distance tests using
        the Gaussian kernel with bandwidth parameter h.
        Depending on the shape of the `y`, the function performs the tests of
        multivariate normality, the non-parametric two-sample tests or the k-sample tests.

        Parameters
        ----------
            x : numpy.ndarray or pandas.DataFrame.
                A numeric array of data values.
            y : numpy.ndarray or pandas.DataFrame, optional
                A numeric array data values (for two-sample test) and a 1D array of class labels
                (for k-sample test). Defaults to None.

        Returns
        -------
            self : object
                Fitted estimator.
        """

        self.x = x
        self.y = y

        if isinstance(self.x, (np.ndarray, pd.Series)):
            if self.x.ndim == 1:
                self.x = np.array(self.x).reshape(-1, 1)
        elif isinstance(self.x, pd.DataFrame):
            self.x = self.x.to_numpy()
        else:
            raise TypeError("x must be a numpy array or a pandas dataframe")

        if self.y is not None:
            if isinstance(self.y, (np.ndarray, pd.Series)):
                if self.y.ndim == 1:
                    self.y = np.array(self.y).reshape(-1, 1)
            elif isinstance(y, pd.DataFrame):
                self.y = self.y.to_numpy()
            else:
                raise TypeError("y must be a numpy array or a pandas dataframe")

        valid_methods = ["bootstrap", "permutation", "subsampling"]
        if self.method not in valid_methods:
            raise ValueError(
                "method must be one of 'bootstrap', 'permutation', or 'subsampling'"
            )

        if self.b is not None:
            if not 0 < self.b <= 1:
                raise ValueError(
                    "b indicates the proportion used for the \
                        subsamples in the subsampling algorithm. It must be in (0, 1]."
                )

        valid_centering_type = ["param", "nonparam"]
        if self.centering_type not in valid_centering_type:
            raise ValueError("centering must be chosen between 'param' and 'nonparam'")

        if self.h is None:
            if self.alternative is None:
                raise ValueError(
                    "You have not specified a value of h. "
                    "Please specify a alternative from 'location', 'scale' or 'skewness' "
                    "for the tuning parameter selection algorithm to run"
                )

        if not isinstance(self.random_state, (int, type(None))):
            raise ValueError("Please specify a integer or None random_state")

        size_x, k = self.x.shape

        if self.y is None:
            if self.h is None:
                self.h = select_h(
                    self.x,
                    y=None,
                    alternative=self.alternative,
                    num_iter=self.num_iter,
                    quantile=self.quantile,
                    n_jobs=self.n_jobs,
                )[0]

            if self.mu_hat is None:
                self.mu_hat = np.zeros(k).reshape(1, -1)
            else:
                self.x = self.x - self.mu_hat
                self.mu_hat = np.zeros(k)

            if self.sigma_hat is None:
                self.sigma_hat = np.eye(k)

            statistic = stat_normality_test(self.x, self.h, self.mu_hat, self.sigma_hat)

            sigma_h = (self.h**2) * np.eye(k)

            var_un = variance_normality_test(sigma_h, self.sigma_hat, size_x)
            cv_un = cv_normality(
                size_x,
                self.h,
                self.mu_hat,
                self.sigma_hat,
                self.num_iter,
                self.quantile,
                self.random_state,
                self.n_jobs,
            ) / np.sqrt(var_un)

            dof, coeff = dof_normality_test(sigma_h, self.sigma_hat)
            qu_q = chi2.ppf(self.quantile, dof)
            cv_vn = coeff * qu_q

            un_h0 = (statistic[0] / np.sqrt(var_un)) > cv_un
            vn_h0 = statistic[1] > cv_vn

            self.test_type_ = "Kernel-based quadratic distance Normality test"
            self.un_h0_rejected_ = un_h0
            self.vn_h0_rejected_ = vn_h0
            self.un_test_statistic_ = statistic[0] / np.sqrt(var_un)
            self.vn_test_statistic_ = statistic[1]
            self.un_cv_ = cv_un
            self.vn_cv_ = cv_vn
            self.var_un_ = var_un
            self.cv_method_ = None
            return self

        else:
            k = len(np.unique(y))
            if k > self.k_threshold:
                if (self.y is not None) and (self.x.shape[1] != self.y.shape[1]):
                    raise ValueError(
                        "'x' and 'y' must have the same number of columns."
                    )

                size_y = self.y.shape[0]
                data_pool = np.vstack((self.x, self.y))

                if self.h is None:
                    self.h = select_h(
                        self.x,
                        self.y,
                        alternative=self.alternative,
                        quantile=self.quantile,
                        method=self.method,
                        num_iter=self.num_iter,
                        n_jobs=self.n_jobs,
                    )[0]

                if self.centering_type == "param":
                    # Compute the estimates of mean and covariance from the data
                    if self.mu_hat is None:
                        self.mu_hat = np.mean(data_pool, axis=0, keepdims=True)
                    if self.sigma_hat is None:
                        self.sigma_hat = np.cov(data_pool, rowvar=False)

                    statistic = stat_two_sample(
                        self.x, self.y, self.h, self.mu_hat, self.sigma_hat, "param"
                    )

                elif self.centering_type == "nonparam":
                    statistic = stat_two_sample(
                        self.x,
                        self.y,
                        self.h,
                        np.zeros(k),
                        np.eye(k),
                        "nonparam",
                    )

                cv = cv_twosample(
                    self.num_iter,
                    self.quantile,
                    data_pool,
                    size_x,
                    size_y,
                    self.h,
                    self.method,
                    self.b,
                    self.random_state,
                    self.n_jobs,
                )

                h0 = statistic[:2] / np.sqrt(statistic[2:]) > cv / np.sqrt(
                    statistic[2:]
                )

                self.test_type_ = "Kernel-based quadratic distance two-sample test"
                self.dn_h0_rejected_, self.trace_h0_rejected_ = h0
                self.dn_test_statistic_, self.trace_test_statistic_ = statistic[
                    :2
                ] / np.sqrt(statistic[2:])
                self.dn_cv_, self.trace_cv_ = cv / np.sqrt(statistic[2:])
                self.cv_method_ = self.method
                self.var_dn_, self.var_trace_ = statistic[2:]
                self.vn_test_statistic_ = None
                self.vn_cv_ = None
                return self

            else:
                if (self.y is not None) and (self.x.shape[0] != self.y.shape[0]):
                    raise ValueError("'x' and 'y' must have the same number of rows.")

                if self.h is None:
                    self.h = select_h(
                        self.x,
                        self.y,
                        alternative=self.alternative,
                        method=self.method,
                        num_iter=self.num_iter,
                        quantile=self.quantile,
                        n_jobs=self.n_jobs,
                    )[0]

                statistic = stat_ksample(self.x, self.y, self.h)
                cv = cv_ksample(
                    self.x,
                    self.y,
                    self.h,
                    self.num_iter,
                    self.b,
                    self.quantile,
                    self.method,
                    self.random_state,
                    self.n_jobs,
                )

                h0 = statistic[:2] / np.sqrt(statistic[2:]) > cv / np.sqrt(
                    statistic[2:]
                )

                self.test_type_ = "Kernel-based quadratic distance K-sample test"
                self.dn_h0_rejected_, self.trace_h0_rejected_ = h0
                self.dn_test_statistic_, self.trace_test_statistic_ = statistic[
                    :2
                ] / np.sqrt(statistic[2:])
                self.dn_cv_, self.trace_cv_ = cv / np.sqrt(statistic[2:])
                self.cv_method_ = self.method
                self.var_dn_, self.var_trace_ = statistic[2:]
                self.vn_test_statistic_ = None
                self.vn_cv_ = None
                return self

    def stats(self):
        """
        Function to generate descriptive statistics per variable (and per group if available).

        Returns
        -------
            summary_stats_df : pandas.DataFrame
                Dataframe of descriptive statistics.
        """
        summary_stats_df = stats(self.x, self.y)
        return summary_stats_df.round(4)

    def summary(self, print_fmt="simple_grid"):
        """
        Summary function generates a table for the kernel test results and the summary statistics.

        Parameters
        ----------
            print_fmt : str, optional.
                Used for printing the output in the desired format. Defaults to "simple_grid".
                Supports all available options in tabulate, see here: https://pypi.org/project/tabulate/.

        Returns
        --------
            summary : str
                A string formatted in the desired output
                format with the kernel test results and summary statistics.
        """

        if self.vn_test_statistic_ is None:

            index_labels = [
                "Test Statistic",
                "Critical Value",
                "H0 is rejected (1 = True, 0 = False)",
            ]
            test_summary = {
                "Dn": [self.dn_test_statistic_, self.dn_cv_, self.dn_h0_rejected_],
                "Trace": [
                    self.trace_test_statistic_,
                    self.trace_cv_,
                    self.trace_h0_rejected_,
                ],
            }
            res = pd.DataFrame(test_summary, index=index_labels)

        else:
            index_labels = [
                "Test Statistic",
                "Critical Value",
                "H0 is rejected (1 = True, 0 = False)",
            ]
            test_summary = {
                "U-Statistic": [
                    self.un_test_statistic_,
                    self.un_cv_,
                    self.un_h0_rejected_,
                ],
                "V-Statistic": [
                    self.vn_test_statistic_,
                    self.vn_cv_,
                    self.vn_h0_rejected_,
                ],
            }
            res = pd.DataFrame(test_summary, index=index_labels)

        summary_stats_df = self.stats()

        if print_fmt == "html":
            summary_string = (
                "Time taken for execution: {} seconds".format(self.execution_time)
                + "<br>Test Results <br>"
                + f"<br> {self.test_type_} <br>"
                + tabulate(res, tablefmt=print_fmt, headers=res.columns)
                + f"<br> CV method: {self.cv_method_} <br>"
                + f"<br> Selected tuning parameter h : {self.h} <br>"
                + "<br> Summary Statistics <br>"
                + tabulate(
                    summary_stats_df,
                    tablefmt=print_fmt,
                    headers=summary_stats_df.columns,
                )
                + "<br>"
            )
        else:
            summary_string = (
                "Time taken for execution: {:.3f} seconds".format(self.execution_time)
                + "\nTest Results \n"
                + f"{self.test_type_} \n"
                + tabulate(res, tablefmt=print_fmt, headers=res.columns)
                + f"\nCV method: {self.cv_method_} \n"
                + f"Selected tuning parameter h : {self.h} \n"
                + "\nSummary Statistics \n"
                + tabulate(
                    summary_stats_df,
                    tablefmt=print_fmt,
                    headers=summary_stats_df.columns,
                )
            )

        return summary_string
